from .date import get_week_number

from arrow import get as arrow_get

import pytest


@pytest.mark.parametrize(
	"date, expected_week_number",
	[
		(arrow_get('2022-01-01'), 1),
		(arrow_get('2022-01-06'), 2),
		(arrow_get('2021-12-31'), 53),
		('', None),
		([1, 2, 3], None),
		(None, None),
	]
)
def test_get_week_number(date, expected_week_number):
	if not hasattr(date, 'strftime'):
		with pytest.raises(AttributeError):
			get_week_number(date)
	else:
		assert get_week_number(date) == expected_week_number



from contextlib import contextmanager
from datetime import datetime, timedelta


# Так, сейчас будет реклама.
# Тебе случалось когда-нибудь обрабатывать большие, ну, массивы каких-то данных или элементов?
# Не по объёму, а количество большое.
# Ну, у меня вот всё время на работе бывает надо обработать сто тыщ строк или миллион.
# Естественно, это занимает кучу времени, поэтому я хочу в процессе видеть, сколько там есть и сколько осталось.
# То есть обычно всё начинается с того, я пишу print(id).
# Потом мне хочется ещё видеть, сколько их там всего, и тогда я пишу print(id, total).
# Потом я понимаю, что у меня вообще-то консоль не только для миллиона id, там вообще-то другие полезные штуки бывают, и я начинаю добавлять \r в конце этих строк, чтоб оно на одном месте было.
# Потом я понимаю, что пришла пора оптимизаций и неплохо бы ещё считать, сколько фиговин в секунду оно успевает делать, поэтому я ещё отмечаю начальное время и каждый раз делю количество на прошедшее время.
# А потом понимаю, что консоль медленная и сам вывод этой херни замедляет процесс, и неплохо бы ещё сделать так, чтобы прогресс обновлялся только раз в ~100 мс.
# Последний раз у меня такой маленький сниппет разросся до 30 строк. Бывало такое?
# Ну а теперь реклама!
# https://asciinema.org/a/UI1aOqjQC1KXx303kaVGrxjQp


# print('==> Doing important stuff...')
# N = 2_000_000
# with polosa(total=N) as p:
# 	for i in range(0, N):
# 		p.tick(i+1, f'real index is {i}')
#		# or
#		#p.tick(caption=f'real index is {i}')
#		# or even
#		#p.tick()
# print('==> All done!')


class _Polosa:
	def __init__(self, total=None, throttle_ms=200):
		self.start = datetime.now()
		self.last_update = None
		self.last_num = 0
		self.throttle_ms = throttle_ms
		self.total = total
		self.buf = ''

	def tick(self, num=None, caption=''):
		now = datetime.now()

		if num == None:
			self.last_num += 1
			num = self.last_num

		self.buf = f'{num}'
		if self.total != None:
			self.buf += f'/{self.total}'
		per_sec = num / (now - self.start).total_seconds()
		self.buf += f'   {per_sec:.1f}/sec   {caption}'

		if (self.last_update != None) and (now - self.last_update < timedelta(milliseconds=self.throttle_ms)):
			return

		print(self.buf, end='   \r')
		self.last_update = now

	def _finalize(self):
		print(self.buf)


@contextmanager
def polosa(*args, **kwargs):
	p = _Polosa(*args, **kwargs)
	try:
		yield p
	finally:
		p._finalize()




from django.conf import settings
from django.contrib.auth.hashers import check_password
from django.contrib.auth.models import User


class StaticBackend:
	"""
	Authenticate against the setting STATIC_AUTH_CREDENTIALS, which is a dict
	of username -> password pairs. Passwords are stored in plaintext
	(so, it's explicitly insecure).
	"""
	def authenticate(self, request, username=None, password=None):
		if username in settings.STATIC_AUTH_CREDENTIALS:
			if settings.STATIC_AUTH_CREDENTIALS[username] == password:
				return self._get_or_create(username)
		return None

	def _get_or_create(self, username):
		try:
			user = User.objects.get(username=username)
		except User.DoesNotExist:
			# Create a new user. There's no need to set a password
			# because only the password from settings.py is checked.
			user = User(username=username)
			# user.is_staff = True
			# user.is_superuser = True
			user.save()
		return user

	def get_user(self, user_id):
		try:
			return User.objects.get(pk=user_id)
		except User.DoesNotExist:
			return None


# https://docs.djangoproject.com/en/2.1/topics/auth/customizing/#writing-an-authentication-backend
class SettingsBackend:
	"""
	Authenticate against the settings ADMIN_LOGIN and ADMIN_PASSWORD.

	Use the login name and a hash of the password. For example:

	ADMIN_LOGIN = 'admin'
	ADMIN_PASSWORD = 'pbkdf2_sha256$30000$Vo0VlMnkR4Bk$qEvtdyZRWTcOsCnI/oQ7fVOu1XAURIZYoOZ3iq8Dr4M='
	"""

	def authenticate(self, request, username=None, password=None):
		login_valid = (settings.ADMIN_LOGIN == username)
		pwd_valid = check_password(password, settings.ADMIN_PASSWORD)
		if login_valid and pwd_valid:
			try:
				user = User.objects.get(username=username)
			except User.DoesNotExist:
				# Create a new user. There's no need to set a password
				# because only the password from settings.py is checked.
				user = User(username=username)
				user.is_staff = True
				user.is_superuser = True
				user.save()
			return user
		return None

	def get_user(self, user_id):
		try:
			return User.objects.get(pk=user_id)
		except User.DoesNotExist:
			return None



from .dict import filter_dict

import pytest


FIRST_SRC_DICT = {
	'Some...': "BODY",
	'once': "told me",
	'the world': "is gonna roll me",
}
SECOND_SRC_DICT = {
	("Some", "body"): ["once", "told me"],
	("The world", "is gonna"): ["roll", "me"],
	("I ain't the", "sharpest tool"): ["in the shed"],
}


@pytest.mark.parametrize(
	"src, leave_only, expected_keys, expected_values",
	[
		(FIRST_SRC_DICT, ["Some...", "once"], ("Some...", "once"), ("BODY", "told me")),
		(FIRST_SRC_DICT, (), (), ()),
		(SECOND_SRC_DICT, [("Some", "body")], (("Some", "body"), ), (["once", "told me"], )),
		(SECOND_SRC_DICT, (), (), ()),
		({}, ("SOME", "BODY"), (), ()),
		(SECOND_SRC_DICT, ("Some...", ), (), ()),
		([], ("SOME", "BODY"), None, None)
	]
)
def test_filter_dict(src, leave_only, expected_keys, expected_values):
	if isinstance(src, dict):
		filtered_dict = filter_dict(src, leave_only)

		assert tuple(filtered_dict.keys()) == expected_keys
		assert tuple(filtered_dict.values()) == expected_values

	else:
		with pytest.raises(AttributeError, match=".* object has no attribute 'items'"):
			filter_dict(src, leave_only)



from .array import as_chunks, ensure_iterable

from hypothesis import given
from hypothesis.strategies import lists, integers


def _collect(x):
	ret = []
	for i in x: ret.append(i)
	return ret


def _flatten(arr):
	ret = []
	for subarr in arr:
		for elem in subarr:
			ret.append(elem)
	return ret


def test_as_chunks():
	assert _collect(as_chunks([], 5)) == []
	assert _collect(as_chunks([1,2,3], 5)) == [[1,2,3]]
	assert _collect(as_chunks([1,2,3], 2)) == [[1,2],[3]]
	assert _collect(as_chunks([1,2,3], 1)) == [[1],[2],[3]]


@given(lists(integers()), integers(min_value=1))
def test_as_chunks_correctness(source, x):
	result = _collect(as_chunks(source, x))
	num_short = 0
	last_short = False
	for e in result:
		assert len(e) <= x
		if len(e) < x:
			num_short += 1
			last_short = True
	assert num_short <= 1
	if num_short > 0:
		assert last_short
	assert(_flatten(result) == source)


@given(lists(integers(), max_size=10), integers(min_value=1, max_value=20))
def test_as_chunks_smallscale(source, x):
	result = _collect(as_chunks(source, x))
	num_short = 0
	last_short = False
	for e in result:
		assert len(e) <= x
		if len(e) < x:
			num_short += 1
			last_short = True
	assert num_short <= 1
	if num_short > 0:
		assert last_short
	assert(_flatten(result) == source)


def test_ensure_iterable():
	assert ensure_iterable(1) == (1,)
	assert ensure_iterable('x') == ('x',)
	assert ensure_iterable(['x']) == ['x']
	assert ensure_iterable(('x', 'y')) == ('x', 'y')



def format_long_list(value, prefix=None, max_items=None):
	ret = ''
	if prefix != None:
		ret = prefix

	leftover = 0
	if len(value) > max_items:
		leftover = len(value) - max_items
		value = value[:max_items]

	ret += ', '.join(value)
	if leftover > 0:
		ret += f' ...and {leftover} more'

	return ret



from ..filterset import FilterSet, create_filter_set as cfs

from collections import namedtuple

from django.forms import Form


Result = namedtuple('Result', ['is_valid', 'errors', 'data'])

def create_filter_set(field_definitions: dict, required=None, parent=None):
	ret = cfs(field_definitions.keys(), required, parent)

	real_fd = field_definitions.copy()
	if parent != None:
		for key in parent._meta['field_definitions']:
			real_fd[key] = parent._meta['field_definitions'][key]

	DynamicFormInner = type('DynamicFormInner', (Form,), real_fd.copy())

	ret._meta = {}
	ret._meta['field_definitions'] = real_fd
	ret._meta['form'] = DynamicFormInner
	return ret

def from_request(FS: FilterSet, request_payload: dict) -> Result:
	fs = FS(**request_payload)

	form_payload = {}
	for f in fs._fields:
		if f in request_payload:
			form_payload[f] = request_payload[f]

	form = FS._meta['form'](form_payload)
	if form.is_valid():
		return Result(is_valid=True, errors=None, data=FS(**form.cleaned_data))
	else:
		return Result(is_valid=False, errors=form.errors, data=None)



from .profiler import before_cursor_execute, after_cursor_execute

import logging; log = logging.getLogger(__name__)

from notalib.time import Timing
from sqlalchemy import create_engine, event
from sqlalchemy.engine import Engine
from django.conf import settings


CLICKHOUSE_URL = settings.CLICKHOUSE_URL
CLICKHOUSE_PROFILE = settings.CLICKHOUSE_PROFILE


# Activate profiler if needed
event.listens_for(Engine, "before_cursor_execute")(before_cursor_execute)
event.listens_for(Engine, "after_cursor_execute")(after_cursor_execute)


# Initialize engine
engine = create_engine(
	CLICKHOUSE_URL,
	pool_size=20,
	pool_pre_ping=False,
	echo=False,
)


def get_connection():
	return engine.connect()


def get_database_name() -> str:
	return engine.url.database


class Query:
	def __init__(self, q=None, **kwargs):
		self.q = q
		self.params = kwargs

	def execute(self):
		conn = get_connection()
		#if CLICKHOUSE_PROFILE: print(str(self.q))
		t = Timing(auto_print=False)
		with t:
			result = conn.execute(self.q, self.params).fetchall()
		return result

	def execute_val(self):
		rows = self.execute()
		assert len(rows) == 1, 'Expected to get exactly one row'
		assert len(rows[0]) == 1, 'Expected to get exactly one value'
		return rows[0][0]

	def execute_list(self):
		rows = self.execute()
		ret = []
		for row in rows:
			ret.append(row[0])
		return ret

	def execute_kv(self):
		rows = self.execute()
		ret = {}
		for row in rows:
			ret[row[0]] = row[1]
		return ret

	def execute_na(self):
		conn = get_connection()
		#if CLICKHOUSE_PROFILE: print(str(self.q))
		t = Timing(auto_print=False)
		with t:
			result = conn.execute(self.q, self.params)
		return result.rowcount



class FilterSet:
	"""
	Don't use this class directly, use create_filter_set instead.
	"""
	# Fields will be initialized by create_filter_set
	_fields = None
	_required = None
	_parent = None
	_meta = None  # for external use
	_frozen = True

	def __init__(self, **kwargs):
		for f in self._required:
			assert (f in kwargs), f"Field `{f}` is mandatory"
		for key in kwargs:
			assert key in self._fields, f"Unknown field `{key}`"
		self.values = kwargs

	def __getattr__(self, name: str):
		if name in self.__dict__:
			return self.__dict__[name]

		assert (not self._frozen), "FilterSet is frozen"
		return self.values[name]

	def unfreeze(self, expected_fields):
		assert sorted(expected_fields) == sorted(self._fields), f"Field set mismatch, expected: {self._fields}"
		self._frozen = False
		return self

	def as_dict(self):
		return self.values.copy()

	def apply(self, handlers):
		"""
		A piece of code that wants to use these filters calls FilterSet.apply()
		with a dict describing how to handle each field:

			qs = Address.objects.all()
			fs.apply({
				'country': lambda value: qs.filter(country=value),
				'city': lambda value: qs.filter(city=value),
			})

		Code MUST provide a handler for every field. If it does not, an exception
		is risen. So, if you ever add a new field, your implementations can't
		just forget to process that field unless you add a handler.

		Every callback is guaranteed to be called only once per apply() call.

		Hint: Making different implementations for one FilterSet is where it shines.
		"""
		assert (not self._frozen), "FilterSet is frozen"
		for field_name in self._fields:
			assert (field_name in handlers), f"FilterSet.apply: no implementation found for `{field_name}`"
			handlers[field_name](field_name, self.values[field_name])


def create_filter_set(fields, required=None, parent=None):
	"""
	FilterSet defines a set of filters that can be applied to some data.
	It's useful on its own, and it's a powerful abstraction for bigger projects.

	1. Definition.

	FilterSet defines some fields that can be used for filtering data.
	They must have valid Python names.

		AddressFilterSet = create_filter_set(['country', 'city'], required=['country'], parent=None)

	FilterSet is chainable. A child inherits all fields of every of its parents.

	2. Instantiation.

	FilterSet accepts kwargs for known field names.

		fs = AddressFilterSet(country='Russia', city='Penza')

	3. Application.

	FilterSet must be unfrozen before the code is allowed to look inside.
	(This method returns self.)

		fs.unfreeze(['country', 'city'])

	If the set of fields does not match FilterSet's set of field, an exception
	is raised.

		print(fs.country)
		print(fs.as_dict())

	---
	Так, представляю тебе новую суперкрутую вещь в pyutil. Называется FilterSet.
	Вот скажи: у тебя бывало такое, что надо в GET- или там POST-запросе принять какие-то параметры и по ним выдать отфильтрованный сет каких-нибудь данных?
	Ну ладно, это, наверное, у всех было.
	А вот бывало такое, что данные абсолютно разные, но фильтры у них одинаковые или по крайней мере имеют общие части?
	Ну типа, допустим, есть у тебя какая-нибудь админка с заказами и отзывами, но и те, и те фильтруются по дате, стране и городу.
	То есть, получается, ты пишешь и там, и там более-менее примерно одинаковый код, который берёт request, вытаскивает из него date, country и city и строит по ним QuerySet с данными.
	Абстрагироваться там как-то всё-таки трудно, потому что модели могут отличаться, ну и сами фильтры разные.
	Вот, а потом ещё иногда случается такое, что добавляется какой-то новый фильтр во все API - ну, например, помимо города ещё и район.
	И в этот момент происходит страшное: большинство реализаций будут просто игнорировать новое поле. Хотя оно важное. Они не просто сломаются, они будут работать, но возвращать не те данные, которые надо. Это вообще пиздец, хуже даже придумать что-то сложно.
	Единственное лекарство - очень долго и муторно отсматривать всё руками в таких случаях.
	Ну или тесты писать, но там та же проблема: надо не забыть покрыть нужными тестами все нужные API, а когда их много, что-нибудь да пропустишь.
	И ты думаешь: вот бы была какая-нибудь такая фигня, чтобы эти фильтры можно было обновить в одном месте и чтобы весь код, который не знает про новое поле, сразу нахуй сломался и его было видно.
	Дамы и господа, представляю вам FilterSet.
	С ним всё становится просто. Берёшь, описываешь чё у вас там в проекте есть:

		DeliveryFilterSet = create_filter_set({
			'area': formplus.IntegerArrayField(required=False, empty_value=None),
			'executor': formplus.StringArrayField(required=False, empty_value=None),
			'date_from': formplus.MonthField(required=False, empty_value=None),
			'date_to': formplus.MonthField(required=False, empty_value=None),
		})

	Если для конкретной вьюхи какие-то ещё параметры помимо глобальных есть, просто берёшь и пишешь об этом:

		@action(detail=False, methods=['get'])
		def stats(self, request):
			fs = create_filter_set({
				'group_by': formplus.ChoiceWithDefault(required=False, default='day', choices=(('day', 'day'), ('month', 'month'))),
			}, parent=DeliveryFilterSet)

	Разумеется, их можно вкладывать на любую глубину.
	Разумеется, библиотека за тебя всё провалидирует и даже сделает машиночитаемые отчёты об ошибках, которые останется просто отдать клиенту:

		parsed = from_request(fs, request.GET)
		if not parsed.is_valid:
			return Response(parsed.errors.as_data(), status=status.HTTP_400_BAD_REQUEST)

	Ну и магия. Перед использованием FilterSet надо разморозить, указав, какие поля ты собираешься оттуда читать. Если код ещё не знает про новое поле - он сразу посылается нахуй:

	data = parsed.data.unfreeze(['area', 'executor', 'date_from', 'date_to', 'group_by']).as_dict()

	После этого остаётся только найти все красные тесты и пофиксить их. Пропустить что-то невозможно. Легко? Невероятно легко.
	Оформляйте предзаказ всего за 99.99$
	"""

	ret_fields = []
	ret_required = []
	ret_parent = parent

	# Copy things from parent if needed
	if parent != None:
		for f in parent._fields:
			ret_fields.append(f)
		for f in parent._required:
			ret_required.append(f)

	for f in fields:
		ret_fields.append(f)
	if required != None:
		for f in required:
			ret_required.append(f)

	class FilterSetInstance(FilterSet):
		_fields = ret_fields
		_required = ret_required
		_parent = ret_parent
	return FilterSetInstance



from .models import Nothing
from .serializers import NothingSerializer

from django.http import HttpResponse
from django.contrib.auth import authenticate, login
from django.views.decorators.csrf import csrf_exempt
from rest_framework import viewsets
from rest_framework.decorators import renderer_classes
from rest_framework.renderers import JSONRenderer


# Example usage:
#
# router.register(r'auth-check', DirtyHackAuthCheckViewSet)
# ...
# path('api/auth', auth_view),


@csrf_exempt
def auth_view(request):
	username = request.POST.get('username', None)
	password = request.POST.get('password', None)
	user = authenticate(request, username=username, password=password)
	if user is not None:
		login(request, user)
		return HttpResponse('{"result": "ok"}', status=200)
	else:
		return HttpResponse('{"result": "fail"}', status=403)


@renderer_classes([JSONRenderer])
class DirtyHackAuthCheckViewSet(viewsets.ReadOnlyModelViewSet):
	queryset = Nothing.objects.none()
	serializer_class = NothingSerializer



from .filterset import create_filter_set

import pytest


def test_filter_set():
	CountryFilterSet = create_filter_set(['country'], required=['country'])
	AddressFilterSet = create_filter_set(['city', 'street'], required=['city'], parent=CountryFilterSet)

	with pytest.raises(AssertionError): CountryFilterSet()
	_ = CountryFilterSet(country='Russia')
	with pytest.raises(AssertionError): AddressFilterSet()
	with pytest.raises(AssertionError): AddressFilterSet(country='Russia')
	_ = AddressFilterSet(country='Russia', city='Penza')

	source = {
		'country': 'Russia',
		'city': 'Penza',
		'street': 'Something st.',
	}
	fs = AddressFilterSet(**source)

	for key in source:
		with pytest.raises(AssertionError):
			getattr(fs, key)

	with pytest.raises(AssertionError): fs.unfreeze([])
	with pytest.raises(AssertionError): fs.unfreeze(['country'])
	with pytest.raises(AssertionError): fs.unfreeze(['country', 'city'])

	fs.unfreeze(['country', 'city', 'street'])

	assert fs.as_dict() == source

	for key in source:
		assert getattr(fs, key) == source[key]

	### apply

	def do_nothing(*args, **kwargs): pass

	with pytest.raises(AssertionError): fs.apply({})
	with pytest.raises(AssertionError): fs.apply({'country': do_nothing})
	with pytest.raises(AssertionError): fs.apply({'country': do_nothing, 'city': do_nothing})

	class SomeMock:
		called_times = 0
		def assert_country(self, value):
			assert value == 'Russia'
			self.called_times += 1
		def assert_city(self, value):
			assert value == 'Penza'
			self.called_times += 1
		def assert_street(self, value):
			assert value == 'Something st.'
			self.called_times += 1
	tfa = SomeMock()

	fs.apply({
		'country': lambda f, value: tfa.assert_country(value),
		'city': lambda f, value: tfa.assert_city(value),
		'street': lambda f, value: tfa.assert_street(value),
	})
	assert tfa.called_times == 3



from .array import ensure_iterable

import arrow


def parse_month(yyyy_mm):
	a = arrow.get(yyyy_mm, 'YYYY-M')
	return (a.year, a.month)


def parse_date(s, format_or_formats):
	d = None
	for f in ensure_iterable(format_or_formats):
		try:
			d = arrow.get(s, f)
			break
		except:  # should probably narrow to arrow.ParserError
			pass
	if d == None:
		raise ValueError(f"Could not parse date `{s}` with any of the specified formats")
	return d


def normalize_date(s, input_formats, output_format, allow_empty=True):
	if s == None and allow_empty:
		return None
	return parse_date(s, input_formats).format(output_format)


def get_week_number(date: arrow.Arrow) -> int:
	"""
	Returns the number of the week by date

	Notes
	-----
	strftime method with parameter '%W' returns week number in range 0...53 where is a monday is a first day of week.

	01.01.2022 is a first week of year, but method returns 0. To result in a human-readable format, 1 is added.
	"""

	return int(date.strftime('%W')) + 1



from html.parser import HTMLParser

import re

REGEX = re.compile('<[^>]*>')

class MLStripper(HTMLParser):
	def __init__(self):
		super().__init__()
		self.reset()
		self.strict = False
		self.convert_charrefs = True
		self.fed = []
	def handle_data(self, d):
		self.fed.append(d)
	def get_data(self):
		return ' '.join(self.fed)

def strip_tags(s, fast_and_dirty=False):
	if fast_and_dirty:
		return REGEX.sub(' ', s)
	else:
		p = MLStripper()
		p.feed(s)
		return p.get_data()


class TablePrinter():
	"""
	Prints an HTML table, row by row, from the given data, using attrs or
	dictionary keys as columns.

	Two ways to use it:

	* Call header() / entry() / footer() manually
	
	```
	>>> from notalib.hypertext import TablePrinter
	>>> t = TablePrinter(['a', 'b'])
	>>> t.header()
	'<table><thead><tr><th>a</th><th>b</th></tr></thead><tbody>'
	>>> t.entry({'a': 1, 'b': 2})
	'<tr><td>1</td><td>2</td></tr>\n'
	>>> t.entry({'a': 11, 'b': 22})
	'<tr><td>11</td><td>22</td></tr>\n'
	>>> t.footer()
	'</tbody></table>'
	```

	* Pass an iterable to iterator_over()
	
	```
	>>> from notalib.hypertext import TablePrinter
	>>> t = TablePrinter(['a', 'b'])
	>>> list(t.iterator_over([ {'a': 11, 'b': 22} ]))
	['<table><thead><tr><th>a</th><th>b</th></tr></thead><tbody>', '<tr><td>11</td><td>22</td></tr>\n', '</tbody></table>']
	```
	"""
	def __init__(self, cols=(), use_attrs=False):
		self.cols = cols
		self._get = getattr if use_attrs else lambda arr, i: arr[i]

	def header(self): return '<table><thead><tr>' + ''.join(f'<th>{col}</th>' for col in self.cols) + '</tr></thead><tbody>'
	def entry(self, x): return '<tr>' + ''.join(f'<td>{self._get(x, col)}</td>' for col in self.cols) + '</tr>\n'
	def footer(self): return '</tbody></table>'

	def iterator_over(self, collection, data_only=False):
		if not data_only: yield self.header()
		for i in collection: yield self.entry(i)
		if not data_only: yield self.footer()



class Range(object):
	"""
	Unlike most libraries (and the standard library), this one works with most
	generic definition of range. Range is defined as just anything that
	has start and end. The only requirement is for these values to be comparable.
	So, this works with ints, timestamps, tuples, whatever.

	https://stackoverflow.com/a/48265052
	"""
	def __init__(self, start, end):
		self.start = start
		self.end = end
		# self.duration = self.end - self.start

	def is_overlapped(self, other_range):
		if max(self.start, other_range.start) < min(self.end, other_range.end):
			return True
		else:
			return False

	def get_overlapped_range(self, other_range):
		if not self.is_overlapped(other_range): return
		if other_range.start >= self.start:
			if self.end >= other_range.end:
				return Range(other_range.start, other_range.end)
			else:
				return Range(other_range.start, self.end)
		elif other_range.start < self.start:
			if other_range.end >= self.end:
				return Range(self.start, self.end)
			else:
				return Range(self.start, other_range.end)

	# def __repr__(self):
	# 	return '{0} ------> {1}'.format(*[time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(d))
	# 		for d in [self.start, self.end]])



from .stream import stream_json

from typing import Iterable

import pytest
from django.http import StreamingHttpResponse


@pytest.mark.parametrize(
	"data, expected_data",
	[
		([{'1': '2'}, {'1': '3'}], b'[{"1": "2"}, {"1": "3"}]'),
		(({'1': '2'}, {'1': '3'}), b'[{"1": "2"}, {"1": "3"}]'),
		(map(lambda x: x, [{'1': '2'}, {'1': '3'}]), b'[{"1": "2"}, {"1": "3"}]'),
		('HelloWorld!', b'["H", "e", "l", "l", "o", "W", "o", "r", "l", "d", "!"]'),
		(1511, None),
		([], b'[]'),
		((), b'[]'),
	]
)
def test_stream_json(data, expected_data):
	if isinstance(data, Iterable):
		res = stream_json(data)
		assert isinstance(res, StreamingHttpResponse)
		assert res.getvalue() == expected_data
	else:
		with pytest.raises(AssertionError, match="stream_json: data must be of iterable type"):
			stream_json(data)



def as_chunks(arr, n):
	"""
	Yield successive n-sized chunks from arr.
	"""
	assert n >= 1
	for i in range(0, len(arr), n):
		yield arr[i:i + n]


def ensure_iterable(x):
	if isinstance(x, list): return x
	if isinstance(x, tuple): return x
	# TODO other cases
	return (x,)



from .pandasplus import replace_null_objects

from pytest import mark
from arrow import get as arrow_get
from pandas import DataFrame, NaT, NA


SOURCE_DATAFRAME = DataFrame(
	{
		"dates": [arrow_get("2021-09-07").date(), arrow_get("2007-01-01").date(), NaT],
		"strings": ["Some Data", None, NA],
		"numbers": [1511, NA, 0],
	}
)


@mark.parametrize(
	"src, new_value, expected_proc",
	[
		(
			SOURCE_DATAFRAME,
			None,
			{
				"dates": [arrow_get("2021-09-07").date(), arrow_get("2007-01-01").date(), None],
				"strings": ["Some Data", None, None],
				"numbers": [1511, None, 0],
			},
		),
		(
			SOURCE_DATAFRAME,
			"notalib but a conservative",
			{
				"dates": [arrow_get("2021-09-07").date(), arrow_get("2007-01-01").date(), "notalib but a conservative"],
				"strings": ["Some Data", "notalib but a conservative", "notalib but a conservative"],
				"numbers": [1511, "notalib but a conservative", 0],
			},
		),
		(
			SOURCE_DATAFRAME,
			1998,
			{
				"dates": [arrow_get("2021-09-07").date(), arrow_get("2007-01-01").date(), 1998],
				"strings": ["Some Data", 1998, 1998],
				"numbers": [1511, 1998, 0],
			},
		),
		(
			DataFrame(),
			None,
			{},
		)
	]
)
def test_replace_null_objects(src, new_value, expected_proc):
	proc = replace_null_objects(src, new_value)

	if not src.empty:
		assert proc.columns.tolist() == list(expected_proc.keys())

		assert proc.dates.tolist() == expected_proc['dates']
		assert proc.strings.tolist() == expected_proc['strings']
		assert proc.numbers.tolist() == expected_proc['numbers']

	else:
		assert proc.empty
